﻿using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Windows;
using System.Windows.Data;

using MindFusion.Diagramming.Wpf;
using MindFusion.Diagramming.Wpf.Layout;


namespace VirtualDiagLayout
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window
	{
		public MainWindow()
		{
			InitializeComponent();
		}

		void OnLoaded(object sender, RoutedEventArgs e)
		{
			ConfigDiagram();
			CreateVirtualItems();
			ArrangeContainers();

			diagram.NodesSource = nodes;
			diagram.LinksSource = links;
		}

		void ArrangeContainers()
		{
			var virtualDiagram = new VirtualizingDiagram();
			virtualDiagram.InstantiateNode =
				(dataItem) => InstantiateNode(virtualDiagram, dataItem);
			virtualDiagram.NodesSource = nodes;
			virtualDiagram.LinksSource = links;

			// this creates actual nodes and links for all data items
			var inifiniteViewport = new Rect(0, 0, 1000000, 1000000);
			virtualDiagram.RealizeItems(inifiniteViewport);

			// container.Arrange does not work for VirtualizingDiagram
			var offscreenDiagram = new Diagram();
			virtualDiagram.CopyTo(offscreenDiagram);

			// apply layout to containers
			var containers = offscreenDiagram.Nodes.OfType<ContainerNode>();
			foreach (var ctr in containers)
				ctr.Arrange(new TreeLayout { LinkStyle = TreeLayoutLinkType.Cascading3 });

			// apply layout to diagram
			offscreenDiagram.Arrange(
				new LayeredLayout { MultipleGraphsPlacement = MultipleGraphsPlacement.Horizontal });

			// now actual items have coordinates
			for (int i = 0; i < offscreenDiagram.Nodes.Count; i++)
			{
				var arrangedNode = offscreenDiagram.Nodes[i];
				var rect = arrangedNode.Bounds;

				var virtualNode = virtualDiagram.Nodes[i];
				var nodeData = (NodeData)virtualNode.DataContext;
				nodeData.X = rect.X;
				nodeData.Y = rect.Y;
				nodeData.Width = rect.Width;
				nodeData.Height = rect.Height;
			}
			for (int i = 0; i < offscreenDiagram.Links.Count; i++)
			{
				var arrangedLink = offscreenDiagram.Links[i];
				var points = arrangedLink.Bounds;

				var virtualLink = virtualDiagram.Links[i];
				var linkData = (LinkData)virtualLink.DataContext;
				linkData.ControlPoints = arrangedLink.ControlPoints;
			}
		}

		ObservableCollection<NodeData> nodes = new();
		ObservableCollection<LinkData> links = new();

		void CreateVirtualItems()
		{
			for (int i = 0; i < 30; i++)
			{
				// container data item
				var ctr = new NodeData();
				ctr.Value = i;
				ctr.Text = i.ToString();
				ctr.NodeType = "C";
				nodes.Add(ctr);

				// child data items
				List<NodeData> children = new();
				for (int j = 0; j < 3; ++j)
				{
					var node = new NodeData();
					node.Value = j;
					node.Text = j.ToString();
					node.NodeType = "S";
					node.Container = ctr;
					nodes.Add(node);
					children.Add(node);
				}

				// link data items
				for (int j = 1; j <= 2; ++j)
				{
					var link = new LinkData();
					link.Origin = children[0];
					link.Destination = children[j];
					link.Text = "";
					links.Add(link);
				}
			}
		}

		void ConfigDiagram()
		{
			diagram.Bounds = new Rect(0, 0, 100000, 1520);
			diagram.DefaultNodeSize = new Size(50, 60);
			diagram.BindingMode = BindingMode.TwoWay;

			diagram.InstantiateNode =
				(dataItem) => InstantiateNode(diagram, dataItem);
			diagram.LinkRealized += OnLinkRealized;
		}

		void OnLinkRealized(object? sender, VirtualLinkEventArgs e)
		{
			var linkData = (LinkData)e.DataItem;
			e.Link.ControlPoints.Clear();
			e.Link.ControlPoints.AddRange(linkData.ControlPoints);
			e.Link.UpdateFromPoints();
		}

		DiagramNode InstantiateNode(Diagram targetDiagram, object dataItem)
		{
			var nd = (NodeData)dataItem;
			DiagramNode node = null;
			switch (nd.NodeType)
			{
				case "S":
					node = new ShapeNode();
					break;
				case "C":
					node = new ContainerNode();
					break;
			}
			targetDiagram.Nodes.Add(node);
			node.Bounds = new Rect(
				nd.X, nd.Y, nd.Width, nd.Height);
			return node;
		}
	}

	public class NodeData : INotifyPropertyChanged
	{
		public NodeData Container { get; set; }
		public string NodeType;

		private double x;
		public double X
		{
			get { return x; }
			set
			{
				if (x != value)
				{
					x = value;
					PropertyChanged?.Invoke(
						this, new PropertyChangedEventArgs("X"));
				}
			}
		}

		private double y;
		public double Y
		{
			get { return y; }
			set
			{
				if (y != value)
				{
					y = value;
					PropertyChanged?.Invoke(
						this, new PropertyChangedEventArgs("Y"));
				}
			}
		}

		private double width = 50;
		public double Width
		{
			get { return width; }
			set
			{
				if (width != value)
				{
					width = value;
					PropertyChanged?.Invoke(
						this, new PropertyChangedEventArgs("Width"));
				}
			}
		}

		private double height = 50;
		public double Height
		{
			get { return height; }
			set
			{
				if (height != value)
				{
					height = value;
					PropertyChanged?.Invoke(
						this, new PropertyChangedEventArgs("Height"));
				}
			}
		}

		private string text;
		public string Text
		{
			get { return text; }
			set
			{
				if (text != value)
				{
					text = value;
					PropertyChanged?.Invoke(
						this, new PropertyChangedEventArgs("Text"));
				}
			}
		}

		public int Value { get; set; }
		public event PropertyChangedEventHandler PropertyChanged;
	}

	public class LinkData : INotifyPropertyChanged
	{
		public PointCollection ControlPoints { get; set; }

		private NodeData origin;
		public NodeData Origin
		{
			get { return origin; }
			set
			{
				if (origin != value)
				{
					origin = value;
					PropertyChanged?.Invoke(
						this, new PropertyChangedEventArgs("Origin"));
				}
			}
		}

		private NodeData destination;
		public NodeData Destination
		{
			get { return destination; }
			set
			{
				if (destination != value)
				{
					destination = value;
					PropertyChanged?.Invoke(
						this, new PropertyChangedEventArgs("Destination"));
				}
			}
		}

		public string Text { get; set; }

		public event PropertyChangedEventHandler PropertyChanged;
	}
}
